@@ -1,4 +1,3 @@
-#!/usr/bin/env python2
 #
 # Copyright 2009 Facebook
 #
@@ -260,9 +259,8 @@ class Template(object):
         :arg str template_string: the contents of the template file.
         :arg str name: the filename from which the template was loaded
             (used for error message).
-        :arg tornado.template.BaseLoader loader: the `~tornado.template.BaseLoader` responsible for this template,
-            used to resolve ``{% include %}`` and ``{% extend %}``
-            directives.
+        :arg tornado.template.BaseLoader loader: the `~tornado.template.BaseLoader` responsible
+            for this template, used to resolve ``{% include %}`` and ``{% extend %}`` directives.
         :arg bool compress_whitespace: Deprecated since Tornado 4.3.
             Equivalent to ``whitespace="single"`` if true and
             ``whitespace="all"`` if false.
